from typing import Dict, List, Literal, Optional, Tuple, Union

import numpy as np
from pydantic_ome_ngff.v04 import Axis
from pydantic_ome_ngff.v04.multiscale import Dataset, MultiscaleMetadata
from pydantic_ome_ngff.v04.transform import VectorScale
from skimage.transform import downscale_local_mean, rescale

from mrc2omezarr.header import MrcHeader


def convert(
    data: np.ndarray,
    header: MrcHeader,
    scale_factors: List[Dict[str, int]],
    voxel_size: Optional[Union[float, Tuple[float, float, float]]] = None,
    pyramid_method: Union[Literal["local_mean"], Literal["downsample"]] = "local_mean",
):
    if voxel_size is None:
        voxel_size = header.voxel_size

    if isinstance(voxel_size, float):
        voxel_size = (voxel_size, voxel_size, voxel_size)

    data_pyramid = []
    meta = []

    for idx, sf in enumerate(scale_factors):
        scale = VectorScale(scale=[sf["z"] * voxel_size[0], sf["y"] * voxel_size[1], sf["x"] * voxel_size[2]])
        ms = Dataset(path=f"{idx}", coordinateTransformations=[scale])
        meta.append(ms)
        if pyramid_method == "local_mean":
            data_pyramid.append(downscale_local_mean(data, (sf["z"], sf["y"], sf["x"])))
        elif pyramid_method == "downsample":
            data_pyramid.append(
                rescale(
                    data,
                    (1 / sf["z"], 1 / sf["y"], 1 / sf["x"]),
                    anti_aliasing=False,
                    preserve_range=True,
                    order=0,
                ),
            )

    meta_pyramid = MultiscaleMetadata(
        name="/",
        axes=[
            Axis(name="z", type="space", unit="angstrom"),
            Axis(name="y", type="space", unit="angstrom"),
            Axis(name="x", type="space", unit="angstrom"),
        ],
        metadata={"source_mrc_header": header.model_dump()},
        datasets=tuple(meta),
    )

    return data_pyramid, meta_pyramid


def convert_array(
    data: np.ndarray,
    scale_factors: Tuple[int],
    voxel_size: Union[float, Tuple[float, float, float]],
    pyramid_method: Union[Literal["local_mean"], Literal["downsample"]] = "local_mean",
    is_image_stack: bool = False,
):
    scale_maps = []
    for sf in scale_factors:
        scale_maps.append(
            {
                "x": sf,
                "y": sf,
                "z": 1 if is_image_stack else sf,
            },
        )

    scale_factors = scale_maps

    if isinstance(voxel_size, float):
        voxel_size = (voxel_size, voxel_size, voxel_size)

    data_pyramid = []
    meta = []

    for idx, sf in enumerate(scale_factors):
        scale = VectorScale(scale=[sf["z"] * voxel_size[0], sf["y"] * voxel_size[1], sf["x"] * voxel_size[2]])
        ms = Dataset(path=f"{idx}", coordinateTransformations=[scale])
        meta.append(ms)
        if pyramid_method == "local_mean":
            data_pyramid.append(downscale_local_mean(data, (sf["z"], sf["y"], sf["x"])))
        elif pyramid_method == "downsample":
            data_pyramid.append(
                rescale(
                    data,
                    (1 / sf["z"], 1 / sf["y"], 1 / sf["x"]),
                    anti_aliasing=False,
                    preserve_range=True,
                    order=0,
                ),
            )

    meta_pyramid = MultiscaleMetadata(
        name="/",
        axes=[
            Axis(name="z", type="space", unit="angstrom"),
            Axis(name="y", type="space", unit="angstrom"),
            Axis(name="x", type="space", unit="angstrom"),
        ],
        datasets=tuple(meta),
        metadata={},
    )

    return data_pyramid, meta_pyramid



import json
from typing import Any, Dict, Tuple


def get_protocol(path: str) -> Tuple[str, str]:
    parts = path.split("://")

    if len(parts) == 1:
        return "local", path

    return parts[0], parts[1]


def get_filesystem_args(path: str) -> Dict[str, Any]:
    with open(path) as f:
        return json.load(f)



from typing import List, Tuple

import fsspec
import numpy as np
import ome_zarr.writer
import zarr
from pydantic_ome_ngff.v04 import MultiscaleMetadata

from mrc2omezarr.util import get_protocol


class Writer:
    def __init__(self, data: List[np.ndarray], meta: MultiscaleMetadata) -> None:
        self.data = data
        self.meta = meta

    def write(
        self,
        output_path: str,
        overwrite: bool = True,
        chunks: Tuple[int, int, int] = (256, 256, 256),
        **kwargs,
    ) -> None:
        protocol, path = get_protocol(output_path)
        if "auto_mkdir" in kwargs:
            fs = fsspec.filesystem(protocol, **kwargs)
        else:
            fs = fsspec.filesystem(protocol, auto_mkdir=True, **kwargs)

        loc = zarr.storage.FSStore(path, key_separator="/", mode="w", dimension_separator="/", fs=fs)
        root_group = zarr.group(loc, overwrite=overwrite)

        ome_zarr.writer.write_multiscale(
            self.data,
            group=root_group,
            axes=[a.model_dump() for a in self.meta.axes],
            coordinate_transformations=[
                list(mds.model_dump()["coordinateTransformations"]) for mds in self.meta.datasets
            ],
            storage_options=dict(chunks=chunks, overwrite=overwrite),
            compute=True,
            metadata=self.meta.metadata,
        )



from typing import Tuple

import fsspec
import numpy as np
from mrcfile.mrcinterpreter import MrcInterpreter

from mrc2omezarr.header import MrcHeader
from mrc2omezarr.util import get_protocol


class Reader:
    def __init__(self, path: str, **kwargs):
        self.protocol, self.path = get_protocol(path)

        if "auto_mkdir" in kwargs:
            self.fs = fsspec.filesystem(self.protocol, **kwargs)
        else:
            self.fs = fsspec.filesystem(self.protocol, auto_mkdir=True, **kwargs)

    def read(self, permissive: bool = False) -> Tuple[np.ndarray, MrcHeader]:
        with self.fs.open(self.path, "rb") as f:
            mrc = MrcInterpreter(f, permissive=permissive)
            return mrc.data, MrcHeader.from_header(mrc.header)



from typing import List, Tuple

import numpy as np
from pydantic import BaseModel


class CellA(BaseModel):
    x: float
    y: float
    z: float

    @classmethod
    def from_header(cls, header: np.recarray):
        return cls(**{n: header[n] for n in header.dtype.names})


class CellB(BaseModel):
    alpha: float
    beta: float
    gamma: float

    @classmethod
    def from_header(cls, header: np.recarray):
        return cls(**{n: header[n] for n in header.dtype.names})


class Origin(BaseModel):
    x: float
    y: float
    z: float

    @classmethod
    def from_header(cls, header: np.recarray):
        return cls(**{n: header[n] for n in header.dtype.names})


class MrcHeader(BaseModel):
    nx: int
    ny: int
    nz: int

    mode: int

    nxstart: int
    nystart: int
    nzstart: int

    mx: int
    my: int
    mz: int

    cella: CellA
    cellb: CellB

    mapc: int
    mapr: int
    maps: int

    dmin: float
    dmax: float
    dmean: float

    ispg: int
    nsymbt: int

    extra1: str
    exttyp: str
    nversion: int
    extra2: str

    origin: Origin

    map: str
    machst: str

    rms: float

    nlabl: int
    label: List[str]

    @classmethod
    def from_header(cls, header: np.recarray):
        header_dict = {n: header[n] for n in header.dtype.names}

        header_dict["cella"] = CellA.from_header(header["cella"])
        header_dict["cellb"] = CellB.from_header(header["cellb"])

        header_dict["extra1"] = str(header_dict["extra1"].tobytes())
        header_dict["exttyp"] = header["exttyp"].tobytes().decode()
        header_dict["extra2"] = str(header_dict["extra2"].tobytes())

        header_dict["origin"] = Origin.from_header(header["origin"])

        header_dict["map"] = header["map"].tobytes().decode()
        header_dict["machst"] = str(header["machst"].tobytes())
        header_dict["label"] = [lab.decode() for lab in header["label"]]

        return cls(**header_dict)

    @property
    def voxel_size(self) -> Tuple[float, float, float]:
        x = self.cella.x / self.mx
        y = self.cella.y / self.my
        z = self.cella.z / self.mz
        return x, y, z

    @property
    def is_image_stack(self) -> bool:
        return self.nz == 1



from typing import Tuple

import click

from mrc2omezarr.proc import convert_mrc_to_ngff
from mrc2omezarr.util import get_filesystem_args


@click.command(context_settings={"show_default": True})
@click.option(
    "--mrc-path",
    type=str,
    required=True,
    help="Path to the MRC file. Include the protocol if necessary. (e.g. s3://)",
)
@click.option(
    "--zarr-path",
    type=str,
    required=True,
    help="Path to the output Zarr file. Include the protocol if necessary. (e.g. s3://)",
)
@click.option(
    "--permissive/--no-permissive",
    default=False,
    help="Whether to read the MRC file in permissive mode.",
)
@click.option(
    "--overwrite/--no-overwrite",
    default=True,
    help="Whether to overwrite the output Zarr file.",
)
@click.option(
    "--scale-factors",
    type=str,
    default="1,2,4",
    help="Scale factors for multiscale pyramid. Comma-separated list of integers.",
)
@click.option(
    "--voxel-size",
    type=str,
    default=None,
    help="""Voxel size in Angstroms. Comma-separated list of floats or single float. """
    """If not provided, it will be read from the MRC header.""",
)
@click.option(
    "--is-image-stack/--no-is-image-stack",
    default=None,
    help="Whether the data is an image stack (determined from MRC-header by default).",
)
@click.option(
    "--chunk-size",
    type=int,
    default=256,
    help="Chunk size for the Zarr file.",
)
@click.option(
    "--filesystem-args",
    type=str,
    default=None,
    help="Path to a JSON file containing additional arguments to pass to the fsspec-filesystem.",
)
@click.option(
    "--pyramid-method",
    type=str,
    default="local_mean",
    help="Method to downscale the data. Options: local_mean, downsample.",
)
@click.pass_context
def convert(
    ctx,
    mrc_path,
    zarr_path,
    permissive,
    overwrite,
    scale_factors,
    voxel_size,
    is_image_stack,
    chunk_size,
    filesystem_args,
    pyramid_method,
):
    scale_factors: Tuple[int, ...] = tuple([int(x) for x in scale_factors.split(",")])

    if voxel_size is not None:
        voxel_size = [int(x) for x in voxel_size.split(",")]

    if filesystem_args is not None:
        filesystem_args = get_filesystem_args(filesystem_args)

    convert_mrc_to_ngff(
        mrc_path,
        zarr_path,
        permissive,
        overwrite,
        scale_factors,
        voxel_size,
        is_image_stack,
        (chunk_size, chunk_size, chunk_size),
        filesystem_args,
        pyramid_method,
    )


if __name__ == "__main__":
    convert()



from typing import Any, Dict, Literal, Optional, Tuple, Union

from mrc2omezarr.convert import convert
from mrc2omezarr.read import Reader
from mrc2omezarr.write import Writer


def convert_mrc_to_ngff(
    mrc_path: str,
    zarr_path: str,
    permissive: bool = False,
    overwrite: bool = True,
    scale_factors: Tuple[int, ...] = (1, 2, 4),
    voxel_size: Optional[Union[float, Tuple[float, float, float]]] = None,
    is_image_stack: Optional[bool] = None,
    chunk_size: Optional[Tuple[int, int, int]] = (256, 256, 256),
    filesystem_args: Optional[Dict[str, Any]] = None,
    pyramid_method: Union[Literal["local_mean"], Literal["downsample"]] = "local_mean",
) -> None:
    if filesystem_args is None:
        filesystem_args = {}

    reader = Reader(mrc_path)
    data, header = reader.read(permissive=permissive)

    if voxel_size is None:
        voxel_size = header.voxel_size

    if isinstance(voxel_size, float):
        voxel_size = (voxel_size, voxel_size, voxel_size)

    if is_image_stack is None:
        is_image_stack = header.is_image_stack

    # If the data is an image stack, we don't want to scale the z-axis
    # TODO: Image stacks should be saved as image stacks in the first place, not using this workaround
    scale_maps = []
    for sf in scale_factors:
        scale_maps.append(
            {
                "x": sf,
                "y": sf,
                "z": 1 if is_image_stack else sf,
            },
        )

    data_pyramid, meta_pyramid = convert(data, header, scale_maps, voxel_size=voxel_size, pyramid_method=pyramid_method)

    writer = Writer(data_pyramid, meta_pyramid)
    writer.write(zarr_path, overwrite=overwrite, chunks=chunk_size, **filesystem_args)
